#!/usr/bin/env python3
# ------------------------------------------------------------------------------
# Copyright 2023 Open Text.
#
# The only warranties for products and services of Open Text and its
# affiliates and licensors (“Open Text”) are as may be set forth in the
# express warranty statements accompanying such products and services.
# Nothing herein should be construed as constituting an additional
# warranty. Open Text shall not be liable for technical or editorial
# errors or omissions contained herein. The information contained herein
# is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains
# confidential information and a valid license is required for possession,
# use or copying. If this work is provided to the U.S. Government,
# consistent with FAR 12.211 and 12.212, Commercial Computer Software,
# Computer Software Documentation, and Technical Data for Commercial Items
# are licensed to the U.S. Government under vendor's standard commercial
# license.
# ------------------------------------------------------------------------------

import argparse
import sys
import subprocess
import psycopg2 as pg

umc_env_file_path = '/opt/novell/umc/apps/umc-server/prod.env'
    
# Function to extract username, port, and host from the URL
def parse_db_url(db_url):
    # Extract the protocol
    proto = db_url.split('://')[0]

    # Remove the protocol
    url = db_url.replace(proto + '://', '')

    # Extract the user and password (if any)
    userpass, _, url = url.partition('@')
    user, _, encrypted_pass = userpass.partition(':')

    # Extract the host and port
    host, _, url = url.partition('/')
    host, _, port = host.partition(':')
    
    db_name = url

    try:
        decrypted_pwd = subprocess.check_output([
            'python3', '/opt/novell/umc/scripts/umc-config.py',
            'decrypt_db_pwd',
            '--db-username', user,
            '--db-encrypted-pwd', encrypted_pass
        ], universal_newlines=True).strip()
    except subprocess.CalledProcessError:
        sys.exit(1)

    return user, port, host, decrypted_pwd, db_name

def fetch_db_details():
    db_username = db_port = db_host = db_password = db_name = ""
    # Read the environment variable from the file  
    try:
        with open(umc_env_file_path, 'r') as file:
            for line in file:
                if line.startswith('DB_URL='):
                    db_url = line.strip().split('DB_URL=')[1]
                    db_username, db_port, db_host, db_password, db_name = parse_db_url(db_url)
                    break
    except FileNotFoundError:
        sys.exit(1)
    except Exception as e:
        sys.exit(1)
    
    return db_username, db_port, db_host, db_password, db_name

def create_table(args):
    db_username, db_port, db_host, db_password, db_name = fetch_db_details()
    
    try:
        create_command = ["/opt/novell/umc/scripts/umc-config.py", "create_table", "--db-host", f"{db_host}", "--db-port", f"{db_port}", 
                    "--db-username", f"{db_username}", "--db-password", f"{db_password}", "--db-table", f"{args.table_name}", "--db-name", f"{db_name}"]
    
        process = subprocess.Popen(create_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        
        exit_code = process.wait()
        output, error = process.communicate()
        
        output_str = output.decode('utf-8')
        
        error_str = error.decode('utf-8')
        
        if exit_code != 0:
            sys.exit(1)
        else:
            sys.exit(0)
        
    except Exception as e:
        sys.exit(1)
        
def execute_command(db_name, db_command):
    conn = ""
    db_username, db_port, db_host, db_password, umc_db_name = fetch_db_details()
    
    try:
        # Define the database connection parameters
        db_params = {
            'dbname': db_name,
            'user': db_username,
            'password': db_password,
            'host': db_host,
            'port': db_port,
        }
    
        conn = pg.connect(**db_params)
        autocommit = pg.extensions.ISOLATION_LEVEL_AUTOCOMMIT
        conn.set_isolation_level(autocommit)
        cursor = conn.cursor()
        cursor.execute(db_command)
        if "datname" in db_command:
            retval = cursor.fetchone()
        elif "schemaname" in db_command:
            retval = cursor.fetchone()[0]
        else:
            sys.exit(0)
            
        if not retval:
            sys.exit(1)
        else:
            sys.exit(0)
            
    except Exception as e:
        sys.exit(1)
    finally:
        if conn:
            conn.close()


def build_argparser():
    parser = argparse.ArgumentParser(conflict_handler="resolve")
    subparsers = parser.add_subparsers(dest="cmd")
    
    check_db_exist = subparsers.add_parser(
        "check_db_exist",
        help="Check umc db exist in postgres"
    )
    
    create_umc_db = subparsers.add_parser(
        "create_umc_db",
        help="Create umc db in postgres"
    )
    
    check_table_exist = subparsers.add_parser(
        "check_table_exist",
        help="Check table exist in umc"
    )
    
    check_table_exist.add_argument("--table-name")
    
    create_table = subparsers.add_parser(
        "create_table",
        help="Create missing table"
    )
    create_table.add_argument("--table-name")
    
    return parser

if __name__ == "__main__":
    parser = build_argparser()
    args = parser.parse_args()
    
    if not args.cmd:
        parser.print_help()
        sys.exit(1)

    cmd = args.cmd
    
    if cmd == "create_table":
        create_table(args)
    elif cmd == "check_db_exist":
        db_username, db_port, db_host, db_password, umc_db_name = fetch_db_details()
        execute_command("postgres", f"SELECT 1 FROM pg_database WHERE datname = '{umc_db_name}'")
    elif cmd == "create_umc_db":
        db_username, db_port, db_host, db_password, umc_db_name = fetch_db_details()
        execute_command("postgres", f"CREATE DATABASE {umc_db_name}")
    elif cmd == "check_table_exist":
        db_username, db_port, db_host, db_password, umc_db_name = fetch_db_details()
        execute_command(f"{umc_db_name}", f"SELECT EXISTS (SELECT FROM pg_tables WHERE schemaname = 'public' AND tablename  = '{args.table_name}')")