#!/usr/bin/env python3
# ------------------------------------------------------------------------------
# Copyright 2023 - 2024 Open Text.
#
# The only warranties for products and services of Open Text and its
# affiliates and licensors (“Open Text”) are as may be set forth in the
# express warranty statements accompanying such products and services.
# Nothing herein should be construed as constituting an additional
# warranty. Open Text shall not be liable for technical or editorial
# errors or omissions contained herein. The information contained herein
# is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains
# confidential information and a valid license is required for possession,
# use or copying. If this work is provided to the U.S. Government,
# consistent with FAR 12.211 and 12.212, Commercial Computer Software,
# Computer Software Documentation, and Technical Data for Commercial Items
# are licensed to the U.S. Government under vendor's standard commercial
# license.
# ------------------------------------------------------------------------------
import argparse
import re
import psycopg2 as pg
import os
import subprocess
import socket
import sys

UMC_SERVICE_RESTART = '/opt/novell/umc/scripts/restart-services.sh'
umc_env_file_path = '/opt/novell/umc/apps/umc-server/prod.env'
oes_ldap_sysconfig_path = '/etc/sysconfig/novell/oes-ldap'

# List of tables to migrate
tables_to_migrate = ['action', 'tree', 'typeorm_metadata', 'umc_login_sessions', 'umc_migrations', 'user_profile']

source_conn = target_conn = db_username = db_port = db_host = db_password = new_dburl = source_schema = target_schema = ""

def get_localhost_details():
    local_ip, local_hostname, local_fqdn = (
        socket.gethostbyname(socket.gethostname()),
        socket.gethostname(),
        socket.getfqdn(),
    )

    return local_ip, local_hostname, local_fqdn

def alter_table_owner(table_name):
    local_ip, _, local_fqdn = get_localhost_details()
    
    #Modify the owner of table to the actual user if the db is configured locally
    if db_host.lower() == "localhost" or db_host.lower() == local_fqdn.lower():
        alter_table = subprocess.check_output([
            'psql', '-U', 'postgres',
            '-p', f'{db_port}',
            '-d', f'{source_schema}',
            '-c', f'ALTER TABLE "{table_name}" OWNER to "{db_username}";'
        ],).strip()
        print(f"Output for altering owner for table {table_name} for local db is : {alter_table}")
    else:
        print(f"Remote DB is used, If scripts fails make sure to make db user {db_username} as OWNER for the table {table_name} and then run the script again")  
    
def modify_dbuser_rights(db_host, db_user, db_port):
    local_ip, _, local_fqdn = get_localhost_details()

    # Modify the rights for user to create db
    # if it is remote server then admin has to provide the rights create db to user and then retry the script
    if db_host.lower() == "localhost" or db_host.lower() == local_fqdn.lower():
        provide_rights = subprocess.check_output([
            'psql', '-U', 'postgres',
            '-p', f'{db_port}',
            '-c', f'ALTER USER "{db_user}" CREATEDB;'
        ],).strip()
        print(f"Output for the provide rights for local db is : {provide_rights}")
    else:
        print(f"Remote DB is used, If scripts fails make sure to provide CREATEDB role to the db user and then run the script again")

def source_table_check(source_cursor, table_name):
    source_cursor.execute(f"SELECT EXISTS (SELECT FROM pg_tables WHERE schemaname = 'public' AND tablename  = '{table_name}')")
    return source_cursor.fetchone()[0]
    
def target_table_check(target_cursor, table_name):
    target_cursor.execute(f"SELECT EXISTS (SELECT FROM pg_tables WHERE schemaname = 'public' AND tablename  = '{table_name}')")
    return target_cursor.fetchone()[0]
    
def get_cursor(db_user, db_password, db_host, db_port, db_name):
    # Define the database connection parameters
    db_params = {
        'dbname': db_name,
        'user': db_user,
        'password': db_password,
        'host': db_host,
        'port': db_port,
    }
    
    if db_name == source_schema:
        # Connect to the database
        global source_conn
        source_conn = pg.connect(**db_params)
        autocommit = pg.extensions.ISOLATION_LEVEL_AUTOCOMMIT
        source_conn.set_isolation_level(autocommit)
        
        # Create a cursor for executing SQL command
        source_cursor = source_conn.cursor()
        return source_cursor
    elif db_name == target_schema:
        # Connect to the database
        global target_conn
        target_conn = pg.connect(**db_params)
        autocommit = pg.extensions.ISOLATION_LEVEL_AUTOCOMMIT
        target_conn.set_isolation_level(autocommit)
        
        # Create a cursor for executing SQL command
        target_cursor = target_conn.cursor()
        return target_cursor

def close_db_connection():
    if source_conn:
        source_conn.close()
    if target_conn:
        target_conn.close()
        
    
def migrate_tables(db_user, db_host, db_password, db_port):

    try:
        modify_dbuser_rights(db_host, db_user, db_port)
        
        # Get source cursor for executing SQL commands
        source_cursor = get_cursor(db_user, db_password, db_host, db_port, source_schema)

        # Check if the UMC database exists
        source_cursor.execute(f"SELECT 1 FROM pg_database WHERE datname = %s", (target_schema,))
        umc_db_exists = source_cursor.fetchone()

        if not umc_db_exists:
            # If the UMC database doesn't exist, create it
            source_cursor.execute(f'CREATE DATABASE "{target_schema}"')

        # Get target cursor for executing SQL commands
        target_cursor = get_cursor(db_user, db_password, db_host, db_port, target_schema)
        
        os.environ["PGPASSWORD"] = f'{db_password}'
        
        # Iterate through the tables and migrate them if they don't exist in the target schema
        for table_name in tables_to_migrate:
            # Check if the table exists in the source db
            source_table_exist = source_table_check(source_cursor, table_name)
            if source_table_exist:
                target_table_exist = target_table_check(target_cursor, table_name)
                if not target_table_exist:
                    if table_name == "umc_login_sessions" :
                        alter_table_owner(table_name)
    
                    dump_command = [
                        "pg_dump", "-h", db_host, "-U", db_user, "-p", str(db_port), "-t", table_name, source_schema
                    ]
                    psql_command = [
                        "psql", "-U", db_user, "-h", db_host, "-p", str(db_port), "-d", target_schema
                    ]
                    # Run pg_dump and psql separately
                    with subprocess.Popen(dump_command, stdout=subprocess.PIPE) as dump_proc:
                        with subprocess.Popen(psql_command, stdin=dump_proc.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as psql_proc:
                            dump_proc.stdout.close()  # Allow dump_proc to receive a SIGPIPE if psql_proc exits.
                            output, error = psql_proc.communicate()
                            
                    print(f"Migrate Table Output for {table_name}:")
                    print(output.decode())
                    print(f"Migrate Table Error for {table_name}:")
                    print(error.decode())
                    if error.decode():
                        close_db_connection()
                        revert_changes()
                        sys.exit(1)

        print("Tables migrated successfully!")
        modify_db_url()

    except Exception as e:
        print(f"An error occurred while migrating: {str(e)}")
        sys.exit(1)

    finally:
        os.unsetenv("PGPASSWORD")
        close_db_connection()

# Function to delete the table at source    
def check_and_delete_source_table():
    fetch_db_details()
    try:
        # Get source cursor for executing SQL commands
        source_cursor = get_cursor(db_username, db_password, db_host, db_port, source_schema)
        
         # Get target cursor for executing SQL commands
        target_cursor = get_cursor(db_username, db_password, db_host, db_port, source_schema)
        
        # Iterate through the tables and migrate them if they don't exist in the target schema
        for table_name in tables_to_migrate:
            # Check if the table exists in the source db and target db
            source_table_exist = source_table_check(source_cursor, table_name)
            if source_table_exist:
                target_table_exist = target_table_check(target_cursor, table_name)
                if target_table_exist:
                    source_cursor.execute(f"DROP TABLE IF EXISTS {table_name}")
                    
    except Exception as e:
        print(f"An error occurred during deleting the tables in source: {str(e)}")
        sys.exit(1)
    finally:
        close_db_connection()

def modify_db_url():
    old_dburl = ""
    
    # Fetch old db_url 
    with open(umc_env_file_path, "r") as file:
        for line in file:
            if line.startswith("DB_URL="):
                old_dburl = line.strip()                
    
    try:
        # Read the file and replace the old DB_URL with the new DB_URL
        with open(umc_env_file_path, 'r') as file:
            file_data = file.read()
        
        file_data = file_data.replace(old_dburl, new_dburl)
        
        # Write the updated content back to the file
        with open(umc_env_file_path, 'w') as file:
            file.write(file_data)
        
        print(f"DB_URL in {umc_env_file_path} has been updated.")
    except FileNotFoundError:
        print(f"Error: The file {umc_env_file_path} was not found.")
        sys.exit(1)
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        sys.exit(1)
                     
# If any error occured revert all changes 
def revert_changes():
    print(f"In Revert Changes")
    try:
        # Get source cursor for executing SQL commands
        source_cursor = get_cursor(db_username, db_password, db_host, db_port, source_schema)
        # Drop the database
        source_cursor.execute(f'DROP DATABASE "{target_schema}"')
                    
    except Exception as e:
        print(f"An error occurred during reverting the changes: {str(e)}")
        sys.exit(1)
    finally:
        close_db_connection()

# UMC Service restart
def restart_services():
    res = subprocess.run(
        ["sh", UMC_SERVICE_RESTART],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

def find_tree_name():
    try:
        with open(oes_ldap_sysconfig_path, 'r') as file:
            content = file.read()
            match = re.search(r'CONFIG_LDAP_TREE_NAME="([^"]+)"', content)
            if match:
                return match.group(1).lower()
            else:
                return None
    except FileNotFoundError:
        return None

# Function to extract username, port, and host from the URL
def parse_db_url(db_url):
    # Extract the protocol
    proto = db_url.split('://')[0]

    # Remove the protocol
    url = db_url.replace(proto + '://', '')

    # Extract the user and password (if any)
    userpass, _, url = url.partition('@')
    user, _, encrypted_pass = userpass.partition(':')

    # Extract the host and port
    host, _, url = url.partition('/')
    host, _, port = host.partition(':')

    # Extract the path (if any)
    db_name = url
    
    # Calculate the source schema which is already present in DB_URL
    global source_schema
    source_schema = db_name
    
    # Now calculate the target schema as it will be umc_treename
    global tree_name
    if tree_name is None:
        tree_name = find_tree_name()
    
    if tree_name is None:
        print(f'CONFIG_LDAP_TREE_NAME not found in oes-ldap sysconfig file.')
        print(f'Re run the script with --tree_name "tree_name_value" option.')
        print(f'Example -> /opt/novell/umc/scripts/migrate_table.py --tree_name "tree_name_value"')
        sys.exit(1)

    global target_schema
    target_schema = f"umc_{tree_name}"
    
    global new_dburl
    new_dburl = f"DB_URL={proto}://{user}:{encrypted_pass}@{host}:{port}/{target_schema}"

    try:
        decrypted_pwd = subprocess.check_output([
            'python3', '/opt/novell/umc/scripts/umc-config.py',
            'decrypt_db_pwd',
            '--db-username', user,
            '--db-encrypted-pwd', encrypted_pass
        ], universal_newlines=True).strip()
    except subprocess.CalledProcessError:
        print("Error: Failed to decrypt the password.")
        sys.exit(1)

    return user, port, host, decrypted_pwd

def fetch_db_details():
    # Read the environment variable from the file
    global db_username, db_port, db_host, db_password
    
    try:
        with open(umc_env_file_path, 'r') as file:
            for line in file:
                if line.startswith('DB_URL='):
                    db_url = line.strip().split('DB_URL=')[1]
                    db_username, db_port, db_host, db_password = parse_db_url(db_url)
                    break
    except FileNotFoundError:
        print(f"Error: The file {umc_env_file_path} was not found.")
        sys.exit(1)
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        sys.exit(1)

if __name__ == "__main__":
    # Create ArgumentParser object
    parser = argparse.ArgumentParser(description='Script to migrate the tables to umc specific db')
    # Add an optional argument
    parser.add_argument('--tree_name', type=str, help='This argument is optional')
    # Parse the command line arguments
    args = parser.parse_args()
    if args.tree_name is not None:
        tree_name = args.tree_name.lower()
    else:
        tree_name = None
    fetch_db_details()
    migrate_tables(db_username, db_host, db_password, db_port)
    restart_services()
    sys.exit(0)
