#!/usr/bin/env python3
# ------------------------------------------------------------------------------
# Copyright 2023 - 2024 Open Text.
#
# The only warranties for products and services of Open Text and its
# affiliates and licensors (“Open Text”) are as may be set forth in the
# express warranty statements accompanying such products and services.
# Nothing herein should be construed as constituting an additional
# warranty. Open Text shall not be liable for technical or editorial
# errors or omissions contained herein. The information contained herein
# is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains
# confidential information and a valid license is required for possession,
# use or copying. If this work is provided to the U.S. Government,
# consistent with FAR 12.211 and 12.212, Commercial Computer Software,
# Computer Software Documentation, and Technical Data for Commercial Items
# are licensed to the U.S. Government under vendor's standard commercial
# license.
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# (C) Copyright 1995 - 2022 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its
# affiliates and licensors (“Micro Focus”) are as may be set forth in the
# express warranty statements accompanying such products and services.
# Nothing herein should be construed as constituting an additional
# warranty. Micro Focus shall not be liable for technical or editorial
# errors or omissions contained herein. The information contained herein
# is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains
# confidential information and a valid license is required for possession,
# use or copying. If this work is provided to the U.S. Government,
# consistent with FAR 12.211 and 12.212, Commercial Computer Software,
# Computer Software Documentation, and Technical Data for Commercial Items
# are licensed to the U.S. Government under vendor's standard commercial
# license.
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# Copyright 2022 - 2023 Open Text.
#
#  The only warranties for products and services of Open Text and its affiliates and licensors (“Open Text”)
#  are as may be set forth in the express warranty statements accompanying such products and services.
#  Nothing herein should be construed as constituting an additional warranty. Open Text shall not be liable
#  for technical or editorial errors or omissions contained herein. The information contained herein is
#  subject to change without notice.
# ------------------------------------------------------------------------------

import argparse
import os
import json
import random
import socket
import string
import subprocess
import uuid
from urllib.parse import urlparse, quote, unquote

import ldap
import ldap.modlist as modlist
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import psycopg2 as pg

import umcHttpRedirection
import base64
import hashlib
from Crypto.Cipher import AES
from Crypto import Random

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


CHARSET = string.ascii_letters + string.digits
ERROR_CODES = {
    "success": 0,
    "invalid_ldap_details": 1,
    "io_error": 2,
    "invalid_db_creds": 3,
}

DIRECTORY_TREE_NAME = "directoryTreeName"
OSP_JWK_URL = "umcOspJwkUrl"
UMC_JWK_URL = "umcJwkUrl"
ATTRIBUTES_TYPES = "attributeTypes"

UMC_SERVER_CONFIG = "/opt/novell/umc/apps/umc-server/prod.env"
UMC_SYSCONFIG_FILE = "/etc/sysconfig/novell/umc"
OES_REST_CONFIG = "/opt/novell/umc/apps/oes-rest/prod.env"
OSP_PROPERTIES = (
    "/var/opt/novell/tomcat/webapps/osp/config/osp-configuration.properties"
)

PORSTGRES_SETUP = "/opt/novell/umc/scripts/postgres-setup.sh"
UMC_SETUP = "/opt/novell/umc/scripts/umc-setup.sh"

SQL_USER_OUT = "/tmp/user.sql"
SQL_ALTER_TABLE_OUT = "/tmp/table.sql"
SQL_UMC = "/opt/novell/umc/apps/umc-server/init.sql"
SQL_INSERT_TREE ="/tmp/insertTree.sql"

EDIR_CACERT = "/etc/ssl/servercerts/servercert.pem"

def encrypt_db_detail(db_user, db_pwd):
    key = hashlib.sha256(db_user.encode()).digest()
    rem = len(db_pwd) % 16
    padded = str.encode(db_pwd) + (b'\0' * (16 - rem)) if rem > 0 else str.encode(db_pwd)
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(key, AES.MODE_CFB, iv, segment_size=128)
    enc = cipher.encrypt(padded)[:len(db_pwd)]
    return base64.urlsafe_b64encode(iv + enc).decode()

def read_config_file(filename):
    config = {}

    try:
        with open(filename, "r") as file:
            lines = file.readlines()

            for line in lines:
                line = line.strip()
                try:
                    split_index = line.index("=")
                    key = line[:split_index].strip()
                    value = line[split_index + 1 :]

                    config[key] = value

                except ValueError as ve:
                    config[line] = ""
    except (FileNotFoundError, PermissionError) as e:
        config = {}

    return config


def write_config_file(config: dict, filename):
    if not len(config):
        return

    print(f"updating config file: {filename}")
    with open(filename, "w") as file:
        for key in config:
            value = config[key]

            out_line = f"{key}"

            if value:
                out_line += f"={value}"

            file.write(out_line)
            file.write("\n")


def get_common_proxy_creds():
    username_res = subprocess.run(
        ["/opt/novell/proxymgmt/bin/cp_retrieve_proxy_cred", "username"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    password_res = subprocess.run(
        ["/opt/novell/proxymgmt/bin/cp_retrieve_proxy_cred", "password"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    username = username_res.stdout.decode("utf-8").strip()
    password = password_res.stdout.decode("utf-8").strip()

    return username, password


def get_localhost_details():
    local_ip, local_hostname, local_fqdn = (
        socket.gethostbyname(socket.gethostname()),
        socket.gethostname(),
        socket.getfqdn(),
    )

    return local_ip, local_hostname, local_fqdn


def write_config_to_edir(username, password, attrs_to_update):
    def create_object(ldap_client, dn, modlist):
        try:
            ldap_client.add_s(dn, modlist)
        except ldap.LDAPError as e:
            if not isinstance(e, ldap.ALREADY_EXISTS):
                raise e

    def delete_object(ldap_client, dn):
        try:
            ldap_client.delete_s(dn)
        except ldap.LDAPError as e:
            if not isinstance(e, ldap.NO_SUCH_OBJECT):
                raise e

    def modify_object_attr(ldap_client, dn, modlist):
        try:
            ldap_client.modify_s(dn, modlist)
        except ldap.LDAPError as e:
            if not isinstance(e, ldap.TYPE_OR_VALUE_EXISTS):
                raise e

    def create_umc_config_object(ldap_client):
        """
        create and then update the object attrs
        """
        print("creating umc-config-object")
        dn = "cn=umcConfig,cn=Security"
        attrs = {
            "description": b"umc config",
            "objectClass": [b"umcConfig"],
        }

        ldif = modlist.addModlist(attrs)

        try:
            dnexist = ldap_client.search_ext_s(
                dn,
                scope=ldap.SCOPE_BASE,
                attrlist=[OSP_JWK_URL, UMC_JWK_URL],
            )
        except ldap.LDAPError as e:
            dnexist = 0

        if not bool(dnexist):
            create_object(ldap_client, dn, ldif)

        attrs = []
        for attr in attrs_to_update:
            for url in attrs_to_update[attr]:
                attrs.append(
                    (
                        ldap.MOD_ADD,
                        attr,
                        url.encode("utf-8"),
                    )
                )

        modify_object_attr(ldap_client, dn, attrs)

        modify_object_attr(
            ldap_client,
            dn,
            [
                (ldap.MOD_ADD, "ACL", f"7#subtree#{username}#umcJwkUrl".encode("utf-8")),
                (ldap.MOD_ADD, "ACL", f"7#subtree#{username}#umcOspJwkUrl".encode("utf-8")),
                (ldap.MOD_ADD, "ACL", f"3#subtree#[Public]#umcJwkUrl".encode("utf-8")),
                (ldap.MOD_ADD, "ACL", f"3#subtree#[Public]#umcOspJwkUrl".encode("utf-8")),
            ]
        )

    def get_directory_treename(ldap_client):
        treename = ""

        results = ldap_client.search_ext_s(
            "", scope=ldap.SCOPE_BASE, attrlist=[DIRECTORY_TREE_NAME]
        )
        dn, entry = results[0]
        treename = entry["directoryTreeName"][0].decode("utf-8")

        return treename

    treename = None
    local_ip, _, _ = get_localhost_details()
    ldap_client = ldap.initialize(f"ldaps://{local_ip}")

    try:
        ldap_client.simple_bind_s(username, password)

        create_umc_config_object(ldap_client)
        treename = get_directory_treename(ldap_client)

    except (
        ldap.SERVER_DOWN,
        ldap.INVALID_CREDENTIALS,
        ldap.LDAPError,
        FileNotFoundError,
        PermissionError,
    ) as e:
        print(e)
        raise e
    finally:
        ldap_client.unbind()

    return treename


def read_config_from_edir(ldap_server):
    umc_config = {}

    ldap_server_ip = None
    if ldap_server == "localhost":
        ldap_server_ip, _, _ = get_localhost_details()
    else:
        ldap_server_ip = ldap_server

    ldap_client = ldap.initialize(f"ldaps://{ldap_server_ip}")

    try:
        # anonymous bind, reading public attrs
        ldap_client.simple_bind_s("", "")

        results = ldap_client.search_ext_s(
            "cn=umcConfig,cn=Security",
            scope=ldap.SCOPE_BASE,
            attrlist=[OSP_JWK_URL, UMC_JWK_URL],
        )

        for dn, entry in results:
            if OSP_JWK_URL in entry:
                umc_config[OSP_JWK_URL] = list(
                    map(lambda url: url.decode("utf-8"), entry.get(OSP_JWK_URL))
                )

            if UMC_JWK_URL in entry:
                umc_config[UMC_JWK_URL] = list(
                    map(lambda url: url.decode("utf-8"), entry.get(UMC_JWK_URL))
                )

    except (
        ldap.SERVER_DOWN,
        ldap.INVALID_CREDENTIALS,
        ldap.LDAPError,
        FileNotFoundError,
        PermissionError,
    ) as e:
        raise e
    finally:
        ldap_client.unbind()

    return umc_config

# Must be called during upgrade to add the missing entries
def update_conf_file():
    error_code = ERROR_CODES["success"]
    
    try:
        umc_env = read_config_file(UMC_SERVER_CONFIG)
        umc_env["CLUSTER_LISTING_FAILURE_TIMEOUT"] = "2000"
    
        write_config_file(umc_env, UMC_SERVER_CONFIG)
    except (FileNotFoundError, PermissionError) as e:
        print(e)
        error_code = ERROR_CODES["io_error"]

    return error_code 

# This is for updating the oes release in umc server prod.env
def update_oes_version():
    error_code = ERROR_CODES["success"]
    try :
        umc_env = read_config_file(UMC_SERVER_CONFIG)
        umc_env["OES_DISPLAY_RELEASE"] = read_display_release()
        write_config_file(umc_env, UMC_SERVER_CONFIG)
        
    except (FileNotFoundError, PermissionError) as e:
        print(e)
        error_code = ERROR_CODES["io_error"]

    return error_code
        
def update_umc_config(args):
    error_code = ERROR_CODES["success"]
    global tree

    local_ip, local_hostname, local_fqdn = get_localhost_details()
    umc_provider_id = str(uuid.uuid1(clock_seq=random.randint(1, 100)))
    osp_provider_id = str(uuid.uuid1(clock_seq=random.randint(1, 100)))
    db_username = quote(args.db_username)
    db_password = quote(args.db_password)

    db_host = args.db_host.lower() if args.db_host.lower() != "localhost" else local_fqdn

    try:
        jwk_config = {
            "umcJwkUrl": [
                f"{umc_provider_id}=https://{local_ip}/umc-server/api/v1/auth/ldap/jwks"
            ]
        }

        if args.use_osp:
            osp_host, osp_port = args.osp_url.split(":")

            jwk_config["umcOspJwkUrl"] = [
                f"{osp_provider_id}=https://{osp_host}:{osp_port}/osp/s/oes/jwks"
            ]

        treename = write_config_to_edir(
            args.edir_admin_username,
            args.edir_admin_password,
            jwk_config,
        )

        tree = treename

        umc_env = read_config_file(UMC_SERVER_CONFIG)

        umc_env["UMC_INTERFACES"] = f"{local_ip},{local_hostname},{local_fqdn}"

        encrpt_dbPwd = encrypt_db_detail(unquote(db_username), db_password)
        umc_env[
            "DB_URL"
        ] = f"postgres://{db_username}:{encrpt_dbPwd}@{db_host}:{args.db_port}/umc_{tree.lower()}"
        umc_env["DB_TABLE_NAME"] = "umc_login_sessions"

        umc_env["PROVIDER_ID"] = umc_provider_id
        umc_env["PRIVATE_KEY_PATH"] = "/etc/ssl/servercerts/serverkey.pem"
        umc_env["PUBLIC_KEY_PATH"] = "/etc/ssl/servercerts/servercert.pem"
        umc_env["ROOT_CA_STORES"] = "/etc/ssl/servercerts/"

        umc_env["SERVER_LOG_PATH"] = "/var/opt/novell/log/umc/apps/umc-server/server.log"
        umc_env["ERROR_LOG_PATH"] = "/var/opt/novell/log/umc/apps/umc-server/error.log"
        umc_env["DEBUG_LOG_PATH"] = "/var/opt/novell/log/umc/apps/umc-server/debug.log"

        umc_env["USER_ACTIONS_LOG_SUCCESS_ONLY"] = "false"
        
        umc_env["OES_DISPLAY_RELEASE"] = read_display_release()

        if args.use_osp:
            osp_host, osp_port = args.osp_url.split(":")

            umc_env["OSP_INTERFACE"] = osp_host
            umc_env["OSP_INTERFACE_PORT"] = osp_port
            umc_env["OSP_TENANT_ID"] = args.osp_tenant_id
            umc_env["OSP_REDIRECT_URL"] = "/welcome"
            umc_env[
                "OSP_JWK_URL"
            ] = f"{umc_provider_id}=https://{osp_host}:{osp_port}/osp/s/oes/jwks"

        else:
            umc_env["OSP_INTERFACE"] = ""
            umc_env["OSP_INTERFACE_PORT"] = ""
            umc_env["OSP_TENANT_ID"] = ""
            umc_env["OSP_REDIRECT_URL"] = ""
            umc_env["OSP_JWK_URL"] = ""

        umc_env["EDIR_REST_API_PREFIX"] = "/eDirAPI/v1"

        if args.edir_api_port == "9010":
            umc_env["EDIR_REST_API_PORT"] = "9010"
        else:
            umc_env["EDIR_REST_API_PORT"] = args.edir_api_port

        if args.edir_api_server == "localhost":
            umc_env["EDIR_REST_API_SERVER"] = f"{local_ip}"
        else:
            umc_env["EDIR_REST_API_SERVER"] = args.edir_api_server

        if args.use_identity_console:
            ic_host = (
                args.identity_console_host
                if args.identity_console_host != "localhost"
                else local_fqdn
            )
            ic_url = f"https://{ic_host}:{args.identity_console_port}/identityconsole"

            umc_env["IDENTITY_CONSOLE"] = ic_url

        umc_env["CLUSTER_LISTING_FAILURE_TIMEOUT"] = "2000"

        write_config_file(umc_env, UMC_SERVER_CONFIG)

        if args.db_host.lower() == "localhost" or args.db_host.lower() == local_fqdn.lower():
            with open(SQL_USER_OUT, "w") as file:
                file.write(
                    f"CREATE DATABASE umc_{tree.lower()};\n"
                )
                file.write(
                    f"CREATE USER \"{args.db_username}\" WITH PASSWORD '{args.db_password}';\n"
                )
                file.write(
                    f"ALTER USER \"{args.db_username}\" CREATEDB;\n"
                )
            with open(SQL_ALTER_TABLE_OUT, "w") as file:
                file.write(
                    f'GRANT ALL PRIVILEGES ON "umc_login_sessions" to "{args.db_username}";\n'
                )

                file.write(
                    f'GRANT ALL PRIVILEGES ON "action" to "{args.db_username}";\n'
                )

                file.write(
                    f'GRANT ALL PRIVILEGES ON "user_profile" to "{args.db_username}";\n'
                )

                file.write(f'ALTER TABLE "action" OWNER to "{args.db_username}";\n')

                file.write(f'ALTER TABLE "user_profile" OWNER to "{args.db_username}";\n')

                file.write(f'ALTER TABLE "tree" OWNER to "{args.db_username}";\n')
                
                file.write(f'ALTER TABLE "umc_login_sessions" OWNER to "{args.db_username}";\n')

            with open(SQL_INSERT_TREE, "w") as file:
                file.write(
                    f'INSERT INTO "tree"(name, network_interfaces) VALUES (\'{treename}\', ARRAY[\'{local_ip}\']);\n'
                )
        else:
            with open(SQL_USER_OUT, "w") as file:
                file.write(
                    f"CREATE DATABASE umc_{tree.lower()};"
                )

    except (ldap.LDAPError) as e:
        print(e)
        error_code = ERROR_CODES["invalid_ldap_details"]
    except (FileNotFoundError, PermissionError) as e:
        print(e)
        error_code = ERROR_CODES["io_error"]

    return error_code


def update_oes_rest_config(args):
    error_code = ERROR_CODES["success"]

    try:
        oes_rest_env = read_config_file(OES_REST_CONFIG)

        oes_rest_env["ROOT_CA_STORES"] = "/etc/ssl/servercerts/"

        write_config_file(oes_rest_env, OES_REST_CONFIG)
    except (ldap.LDAPError) as e:
        print(e)
        error_code = ERROR_CODES["invalid_ldap_details"]
    except (FileNotFoundError, PermissionError) as e:
        print(e)
        error_code = ERROR_CODES["io_error"]

    return error_code


def get_config(args):
    error_code = ERROR_CODES["success"]

    osp_instances = []
    umc_instances = []
    db_instances = []

    try:
        umc_config = read_config_from_edir(args.ldap_server)

        # parse configured OSP instances
        if umc_config.get(OSP_JWK_URL):
            for jwk_url in umc_config[OSP_JWK_URL]:
                id, url = jwk_url.split("=")
                parsed_url = urlparse(url)
                osp_instances.append(
                    {
                        "host": parsed_url.hostname,
                        "port": str(parsed_url.port) if parsed_url.port else "443",
                    }
                )

        # parse configured UMC instances
        if umc_config.get(UMC_JWK_URL):
            for jwk_url in umc_config[UMC_JWK_URL]:
                id, url = jwk_url.split("=")
                parsed_url = urlparse(url)
                umc_instances.append(
                    {
                        "host": parsed_url.hostname,
                        "port": str(parsed_url.port) if parsed_url.port else "443",
                    }
                )

                try:
                    res = requests.get(
                        f"https://{parsed_url.netloc}/umc-server/api/v1/config",
                        verify=False,
                    )

                    if res.status_code == 200:
                        res_json = res.json()

                        parsed_db_url = urlparse("https://" + res_json.get("dbHost"))

                        db_instances.append(
                            {
                                "host": parsed_db_url.hostname,
                                "port": str(parsed_db_url.port)
                                if parsed_db_url.port
                                else "443",
                            }
                        )
                except requests.exceptions.RequestException as e:
                    continue

        config = {
            "umc_instances": umc_instances,
            "osp_instances": osp_instances,
            "db_instances": db_instances,
        }
        print(json.dumps(config))
    except (ldap.LDAPError) as e:
        print(e)
        error_code = ERROR_CODES["invalid_ldap_details"]

    return error_code


def validate_db(args):
    error_code = ERROR_CODES["success"]

    _, _, local_fqdn = get_localhost_details()
    db_host = args.db_host.lower() if args.db_host.lower() != "localhost" else local_fqdn

    try:
        conn = pg.connect(
            dbname="postgres",
            host=db_host,
            port=args.db_port,
            user=args.db_username,
            password=args.db_password,
            connect_timeout=5,
        )
        cur = conn.cursor()

    except (Exception, pg.OperationalError, pg.Error) as e:
        print(e)
        error_code = ERROR_CODES["invalid_db_creds"]

    return error_code

def decrypt_db_pwd(args):
    private_key = hashlib.sha256(args.db_username.encode()).digest()
    text = base64.urlsafe_b64decode(args.db_encrypted_pwd)
    iv, value = text[:16], text[16:]
    rem = len(value) % 16
    padded = value + (b'\0' * (16 - rem)) if rem > 0 else value
    cipher = AES.new(private_key, AES.MODE_CFB, iv, segment_size=128)
    return (cipher.decrypt(padded)[:len(value)]).decode()

def setup():
    _, _, local_fqdn = get_localhost_details()
    if args.db_host.lower() == "localhost" or args.db_host.lower() == local_fqdn.lower():
        res = subprocess.run(
            ["sh", PORSTGRES_SETUP, "-p", f"{args.db_port}"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        umc_service_override()

def umc_service_override():
    override_directory_path = '/usr/lib/systemd/system/microfocus-umc-server.service.d/'
    override_file_path = os.path.join(override_directory_path, 'microfocus-umc-server-ovr.conf')
    content = '[Unit]\nWants=postgresql.service\n'

    try:

        # Check if the directory exists, create it if not
        if not os.path.exists(override_directory_path):
            os.makedirs(override_directory_path)
            print(f"Directory '{override_directory_path}' created successfully.")
        
        # Open the file in write mode
        with open(override_file_path, 'w') as file:
            # Write the content to the file
            file.write(content)
            print(f"File '{override_file_path}' created successfully.")
    except IOError as e:
        print(f"Error creating/writing to the file '{override_file_path}': {e}")

def init_db(args):
    error_code = ERROR_CODES["success"]
    local_ip, _, local_fqdn = get_localhost_details()

    if args.db_host.lower() == "localhost" or args.db_host.lower() == local_fqdn.lower():
        # execute tmp script to create a role
        res = subprocess.run(
            ["psql", "-U", "postgres", "-p", f"{args.db_port}", "-f", SQL_USER_OUT],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # create required tables
        res = subprocess.run(
            ["psql", "-U", "postgres", "-d", f"umc_{tree.lower()}", "-p", f"{args.db_port}", "-f", SQL_UMC],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # enter tree details
        res = subprocess.run(
            ["psql", "-U", "postgres", "-d", f"umc_{tree.lower()}", "-p", f"{args.db_port}", "-f", SQL_INSERT_TREE],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # execute tmp script to assign privileges
        res = subprocess.run(
            ["psql", "-U", "postgres", "-d", f"umc_{tree.lower()}", "-p", f"{args.db_port}", "-f", SQL_ALTER_TABLE_OUT],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # delete tmp scripts
        res = subprocess.run(
            ["rm", SQL_USER_OUT, SQL_ALTER_TABLE_OUT, SQL_INSERT_TREE],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    else:
        error_code = execute_command(args, SQL_USER_OUT, "postgres")
        error_code = execute_command(args, SQL_UMC, f"umc_{tree.lower()}")

        res = subprocess.run(
            ["rm", SQL_USER_OUT],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    return error_code

def create_table(args):
    if args.db_table == "action":
        SQL_CREATE_TABLE = "/opt/novell/umc/apps/umc-server/action.sql"
    elif args.db_table == "tree":
        SQL_CREATE_TABLE = "/opt/novell/umc/apps/umc-server/tree.sql"
    elif args.db_table == "umc_login_sessions":
        SQL_CREATE_TABLE = "/opt/novell/umc/apps/umc-server/login_session.sql"
    elif args.db_table == "user_profile":
        SQL_CREATE_TABLE = "/opt/novell/umc/apps/umc-server/user_profile.sql"

    error_code = execute_command(args, SQL_CREATE_TABLE, args.db_name)

    return error_code

def execute_command(args, SQL_CMD, DB_NAME):
    error_code = ERROR_CODES["success"]
    local_ip, _, _ = get_localhost_details()

    try:
        conn = pg.connect(
           dbname=DB_NAME,
           host=args.db_host.lower(),
           port=args.db_port,
           user=args.db_username,
           password=args.db_password,
           connect_timeout=5,
        )
        autocommit = pg.extensions.ISOLATION_LEVEL_AUTOCOMMIT
        conn.set_isolation_level(autocommit)
        cur = conn.cursor()

        # create tables
        with open(SQL_CMD, "r") as file:
            sql = file.read()
            cur.execute(sql)

        if "init" in SQL_CMD:
           cur.execute(
              f"""
              GRANT ALL PRIVILEGES ON "umc_login_sessions" to "{args.db_username}";
              GRANT ALL PRIVILEGES ON "user_profile" to "{args.db_username}";
              GRANT ALL PRIVILEGES ON "action" to "{args.db_username}";
              ALTER TABLE "tree" OWNER to "{args.db_username}";
              INSERT INTO "tree"(name, network_interfaces) VALUES (\'{tree}\', ARRAY[\'{local_ip}\']);
              """
           )
        elif "login_session" in SQL_CMD:
           cur.execute(
              f"""
              GRANT ALL PRIVILEGES ON "umc_login_sessions" to "{args.db_username}";
              """
           )
        elif "action" in SQL_CMD:
           cur.execute(
              f"""
              GRANT ALL PRIVILEGES ON "action" to "{args.db_username}";
              """
           )
        elif "user_profile" in SQL_CMD:
           cur.execute(
              f"""
              GRANT ALL PRIVILEGES ON "user_profile" to "{args.db_username}";
              """
           )
        elif "tree" in SQL_CMD:
           cur.execute(
              f"""
              ALTER TABLE "tree" OWNER to "{args.db_username}";
              """
           )

        conn.commit()
    except (Exception, pg.OperationalError, pg.Error) as e:
        print(e)
        error_code = ERROR_CODES["invalid_db_creds"]

    return error_code

def restart_services():
    res = subprocess.run(
        ["sh", UMC_SETUP],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

def read_display_release():
    file_path = "/etc/novell-release"
    try:
        # Open the file in read mode
        with open(file_path, 'r') as file:
            # Iterate through each line in the file
            for line in file:
                # Check if the line contains DISPLAY_RELEASE
                if "DISPLAY_RELEASE" in line:
                    # Split the line based on the equal sign (=)
                    key, value = map(str.strip, line.split('='))
                    # Return the value of DISPLAY_RELEASE
                    return value
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
    except Exception as e:
        print(f"An error occurred: {e}")

def build_argparser():
    parser = argparse.ArgumentParser(conflict_handler="resolve")
    subparsers = parser.add_subparsers(dest="cmd")

    update_umc_config_group = subparsers.add_parser(
        "update_umc_config",
        help="""writes config to edir (creates schema, attrs, aux class, object) and updates umc-server/prod.env for umc-server
        config object dn: cn=umcConfig,cn=Security
        creates tmp sql files: /tmp/user.sql, /tmp/table.sql scripts (with username and password) that can be invoked to create a ROLE and GRANT privileges
        """,
    )
    update_umc_config_group.add_argument("--edir-admin-username")
    update_umc_config_group.add_argument("--edir-admin-password")

    update_umc_config_group.add_argument("--db-host", default="localhost")
    update_umc_config_group.add_argument("--db-port", default="5432")
    update_umc_config_group.add_argument("--db-username")
    update_umc_config_group.add_argument("--db-password")

    update_umc_config_group.add_argument("--use-osp", action="store_true")
    update_umc_config_group.add_argument("--osp-url")
    update_umc_config_group.add_argument("--osp-tenant-id")
    update_umc_config_group.add_argument("--osp-client-id")

    update_umc_config_group.add_argument("--use-identity-console", action="store_true")
    update_umc_config_group.add_argument("--identity-console-host", default="localhost")
    update_umc_config_group.add_argument("--identity-console-port", default="9000")

    update_umc_config_group.add_argument("--edir-api-server", default="localhost")
    update_umc_config_group.add_argument("--edir-api-port", default="9010")

    oes_rest_config_group = subparsers.add_parser(
        "update_oes_rest_config",
        help="reads the config from edir (JWK urls) and updates oes-rest/prod.env for oes-server",
    )

    oes_rest_config_group.add_argument(
        "--ldap-server",
        default="localhost",
        help="optional param to point to remote edir server",
    )

    get_config_group = subparsers.add_parser(
        "get_config",
        help="reads the config from edir (JWK urls and DB config) and prints it onto the console",
    )

    get_config_group.add_argument(
        "--ldap-server",
        default="localhost",
        help="optional param to point to remote edir server",
    )

    db_validate_group = subparsers.add_parser(
        "validate_db", help="validates the given db parameters"
    )

    db_validate_group.add_argument("--db-host", default="localhost")
    db_validate_group.add_argument("--db-port", default=5432)
    db_validate_group.add_argument("--db-username")
    db_validate_group.add_argument("--db-password")

    db_decrypt_pwd = subparsers.add_parser(
         "decrypt_db_pwd", help="decrypt the db pwd"
    )

    db_decrypt_pwd.add_argument("--db-username")
    db_decrypt_pwd.add_argument("--db-encrypted-pwd")

    create_table = subparsers.add_parser(
         "create_table", help="create the missing table needed by UMC"
    )
    create_table.add_argument("--db-host", default="localhost")
    create_table.add_argument("--db-port", default=5432)
    create_table.add_argument("--db-username")
    create_table.add_argument("--db-password")
    create_table.add_argument("--db-table")
    create_table.add_argument("--db-name")
    
    update_conf_file = subparsers.add_parser(
        "update_conf_file", help="update the entries in conf file, introduced after upgrade"
    )
    
    update_oes_version = subparsers.add_parser(
        "update_oes_version", help="update the version in prod.env for UMC"
    )

    return parser


if __name__ == "__main__":
    ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_ALLOW)
    ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, EDIR_CACERT)

    parser = build_argparser()
    args = parser.parse_args()

    if not args.cmd:
        parser.print_help()
        exit(0)

    cmd = args.cmd

    error_code = 0
    if cmd == "update_umc_config":
        error_code = update_umc_config(args)
        umc_sysconf = read_config_file(UMC_SYSCONFIG_FILE)
        if (
            error_code == ERROR_CODES["success"]
            and umc_sysconf.get("SERVICE_CONFIGURED") != '"yes"'
        ):
            setup()
            init_db(args)
            umcHttpRedirection.main()
        restart_services()
    elif cmd == "update_oes_rest_config":
        error_code = update_oes_rest_config(args)
        umcHttpRedirection.main()
        restart_services()
    elif cmd == "get_config":
        error_code = get_config(args)
    elif cmd == "setup":
        error_code = setup()
    elif cmd == "init_db":
        init_db()
    elif cmd == "validate_db":
        error_code = validate_db(args)
    elif cmd == "create_table":
        error_code = create_table(args)
    elif cmd == "decrypt_db_pwd":
        pwd = decrypt_db_pwd(args)
        print(pwd)
        exit()
    elif cmd == "update_conf_file":
        error_code = update_conf_file()
    elif cmd == "update_oes_version":
        error_code = update_oes_version()
    if error_code:
        print(f"\noperation falied. error_code={error_code}\n")
    else:
        print(f"\nperformed {cmd}\n")

    exit(error_code)
